@@ -8,12 +8,18 @@ import math
 from pathlib import Path
 from typing import Dict, List, Tuple
 
-import h5py
+# import h5py
 import numpy as np
-from h5py import AttributeManager
+from hdf5plugin import Bitshuffle
+from numpy.typing import ArrayLike
+
+# from h5py import AttributeManager
 from scanspec.core import Path as ScanPath
 from scanspec.specs import Line
 
+import h5py  # isort: skip
+from h5py import AttributeManager  # isort: skip
+
 
 def create_attributes(nxs_obj: h5py.Group | h5py.Dataset, names: Tuple, values: Tuple):
     """
@@ -155,7 +161,7 @@ def calculate_scan_range(
     n_images: Tuple | int = None,
     snaked: bool = True,
     rotation: bool = False,
-) -> Dict[str, np.ndarray]:
+) -> Dict[str, ArrayLike]:
     """
     Calculate the scan range for a linear/grid scan or a rotation scan from the number of images to be written.
     If the number of images is not provided, it can be calculated from the increment value of the axis (these values are mutually exclusive).
@@ -179,7 +185,7 @@ def calculate_scan_range(
         ValueError: For a grid scan, if axes_increments is None, n_images must be a tuple of len=2 to be sure to accurately calculate the scan points.
 
     Returns:
-        Dict[str, np.ndarray]: A dictionary of ("axis_name": axis_range) key-value pairs.
+        Dict[str, ArrayLike]: A dictionary of ("axis_name": axis_range) key-value pairs.
     """
     if type(axes_names) != list or type(axes_starts) != list or type(axes_ends) != list:
         raise TypeError("Input values for axes must be passed as lists.")
@@ -203,7 +209,11 @@ def calculate_scan_range(
             n_images = round(abs(axes_starts[0] - axes_ends[0]) / axes_increments[0])
 
         if axes_starts[0] != axes_ends[0] and axes_increments:
-            axes_ends[0] = axes_ends[0] - axes_increments[0]
+            if axes_starts[0] > axes_ends[0]:
+                # Account for reverse rotation.
+                axes_ends[0] = axes_ends[0] + axes_increments[0]
+            else:
+                axes_ends[0] = axes_ends[0] - axes_increments[0]
         elif axes_starts[0] != axes_ends[0] and not axes_increments:
             inc = (axes_ends[0] - axes_starts[0]) / n_images
             axes_ends[0] = axes_ends[0] - inc
@@ -298,18 +308,72 @@ def calculate_origin(
     return det_origin, offset_val
 
 
-def find_number_of_images(datafile_list: List[Path]) -> int:
+def find_number_of_images(datafile_list: List[Path], entry_key: str = "data") -> int:
     """
     Calculate total number of images when there's more than one input HDF5 file.
 
     Args:
         datafile_list (List[Path]): List of paths to the input image files.
+        entry_key (str):    Key for the location of the images inside the data files. Defaults to "data".
 
     Returns:
-        num_images (int): Total number of images._summary_
+        num_images (int): Total number of images.
     """
     num_images = 0
     for filename in datafile_list:
         with h5py.File(filename, "r") as f:
-            num_images += f["data"].shape[0]
+            num_images += f[entry_key].shape[0]
     return int(num_images)
+
+
+# Copy and compress a dataset inside a specified NXclass
+def write_compressed_copy(
+    nxgroup: h5py.Group,
+    dset_name: str,
+    data: ArrayLike = None,
+    filename: Path | str = None,
+    dset_key: str = None,
+    block_size: int = 0,
+):
+    """
+    Write a compressed copy of some dataset in the desired HDF5 group, using the Bitshuffle filter with lz4 compression.
+    The main application for this function in nexgen is to write a compressed copy of a pixel mask or a flatfield file/dataset \
+    directly into the NXdetector group of a NXmx NeXus file.
+    The data and filename arguments are mutually exclusive as only one of them can be used as input.
+    If a filename is passed, it is also required to pass the key for the relevant dataset to be copied. Failure to do so will result \
+    in nothing being written to the NeXus file.
+
+    Args:
+        nxgroup (h5py.Group): Handle to HDF5 group.
+        dset_name (str): Name of the new dataset to be written.
+        data (ArrayLike, optional): Dataset to be compressed. Defaults to None.
+        filename (Path | str, optional): Filename containing the dataset to be compressed into the NeXus file. Defaults to None.
+        dset_key (str, optional): Dataset name inside the passed file. Defaults to None.
+        block_size (int, optional): Number of elements per block, it needs to be divisible by 8. Defaults to 0.
+
+    Raises:
+        ValueError: If both a dataset and a filename have been passed to the function.
+    """
+    from .NXclassWriters import NXclass_logger
+
+    if data is not None and filename is not None:
+        raise ValueError(
+            "The dset and filename arguments are mutually exclusive."
+            "Please pass only the one from which the data should be copied."
+        )
+    if filename and not dset_key:
+        NXclass_logger.warning(
+            f"Missing key to find the dataset to be copied inside {filename}. {dset_name} will not be written into the NeXus file."
+        )
+        return
+
+    if filename:
+        with h5py.File(filename, "r") as fh:
+            data = fh[dset_key][()]
+
+    nxgroup.create_dataset(
+        dset_name, data=data, **Bitshuffle(nelems=block_size, lz4=True)
+    )
+    NXclass_logger.info(
+        f"A compressed copy of the {dset_name} has been written into the NeXus file."
+    )
